/*
** 1998-08-14 -	Configure styles, e.g. things such as back- & foreground colors, icons, and
**		actions. This might become involved.
** 1998-08-23 -	Did lots of work. Now you can actually add and delete styles. Got involved.
** 1998-08-24 -	Completed a first version, fixed plenty of bugs too...
** 1998-08-26 -	Added fun "Copy" button group for visual properties. Also added a deselect
**		signal handler for root style. Oops.
** 1998-08-30 -	Now uses the ico_get_all() function to retrieve a list of all icon file names.
**		Added three (!) forgotten modified-sets. Damn.
** 1998-09-02 -	Added action property configuration, and cleaned up their save/load handling.
** 1998-09-16 -	Made some optimizations, mainly in set_preview() (one more signal blocker).
** 1998-09-27 -	Now uses the brand new cmdseq_dialog for choosing actions. Works fine.
** 1999-03-05 -	Due to new selection handling, we no longer have control over how selected items
**		look. That got rid of some code.
** 1999-03-13 -	Rewrote the dialog module (from scratch), which lead to some changes here too.
** 1999-05-07 -	Now uses the (new) color dialog module. Got rid of lots of crufty code.
** 2000-07-02 -	Took a step closer to the rest of the world, by marking strings for translation.
*/

#include "gentoo.h"

#include "styles.h"
#include "types.h"
#include "dirpane.h"
#include "dialog.h"
#include "iconutil.h"
#include "strutil.h"
#include "fileutil.h"
#include "miscutil.h"
#include "xmlutil.h"
#include "guiutil.h"
#include "style_dialog.h"
#include "cmdseq_dialog.h"
#include "color_dialog.h"
#include "icon_dialog.h"

#include "configure.h"
#include "cfg_module.h"
#include "cfg_paths.h"		/* For cpt_get_path(). */
#include "cfg_types.h"		/* For ctp_get_types(). */
#include "cfg_cmdseq.h"		/* For ccs_get_current(). */

#include "cfg_styles.h"

#define	NODE	"FileStyles"

/* ----------------------------------------------------------------------------------------- */

typedef struct {			/* Visual properties notebook page. */
	GtkWidget	*vbox;
	GtkWidget	*preview;	/* CList showing preview of style (old-school!). */
	GtkWidget	*override[3];	/* Override check buttons for back- & foreground colors, plus icon. */
	GtkWidget	*edit[3];	/* The "edit" (or "pick") command buttons. */
} PVisual;

typedef struct {			/* Action properties notebook page. */
	GtkWidget	*vbox;
	GtkWidget	*actions;	/* List of properties in current style (both local and inherited). */
	GtkWidget	*aname;		/* Entry widget for editing name of an action. */
	GtkWidget	*acmdseq;	/* Entry for command sequence name. */
	GtkWidget	*acmdpick;	/* Pick button for cmdseq. */
	GtkWidget	*adel;		/* Delete (or override) command button. */
} PAction;

typedef struct {
	GtkWidget	*vbox;
	GtkWidget	*scwin;
	GtkWidget	*tree;		/* Main style tree widget. */
	gulong		sig_expand;	/* Signal handler for tree item expansion. */
	gulong		sig_collapse;	/* Signal handler for tree item collapse. */
	GtkTreeStore	*store;

	GtkWidget	*dvbox;		/* Vbox holding definition widgets. */
	GtkWidget	*dname;		/* Name of selected style. */
	GtkWidget	*dparent;	/* Parent of selected style. */
	GtkWidget	*dreparent;	/* Button for reparenting dialog. */

	GtkWidget	*dpnbook;	/* Property notebook. */
	PVisual		dpvisual;
	PAction		dpaction;

	GtkWidget	*del;		/* The style "Delete" button. */

	MainInfo	*min;
	StyleInfo	*si;
	const gchar	*curr_prop;	/* Current property, when editing one. */
	gint		curr_arow;	/* Action property clist row. */
	gboolean	modified;
} P_Styles;

static P_Styles	the_page;

/* ----------------------------------------------------------------------------------------- */

/* 1999-05-24 -	Set the preview widget(s). */
static void set_widgets_preview(P_Styles *page, Style *stl)
{
	const gchar	*iname;
	const GdkColor	*col;

	if((iname = stl_style_property_get_icon(stl, SPN_ICON_UNSEL)) != NULL)
	{
		GdkPixmap	*ipix;
		GdkBitmap	*imsk;

		if((ipix = ico_icon_get(page->min, iname, &imsk)) != NULL)
			gtk_clist_set_pixmap(GTK_CLIST(page->dpvisual.preview), 0, 0, ipix, imsk);
	}
	if((col = stl_style_property_get_color(stl, SPN_COL_UNSEL_BG)) != NULL)
		gtk_clist_set_background(GTK_CLIST(page->dpvisual.preview), 0, (GdkColor *) col);
	else
	{
		col = &gtk_widget_get_style(GTK_WIDGET(page->min->gui->pane[0].list))->base[GTK_STATE_NORMAL];
		gtk_clist_set_background(GTK_CLIST(page->dpvisual.preview), 0, (GdkColor *) col);
	}
	if((col = stl_style_property_get_color(stl, SPN_COL_UNSEL_FG)) != NULL)
		gtk_clist_set_foreground(GTK_CLIST(page->dpvisual.preview), 0, (GdkColor *) col);
}

/* 1999-05-24 -	Set action list. */
static void set_widgets_action(P_Styles *page, Style *stl)
{
	GList	*alist;

	gtk_clist_freeze(GTK_CLIST(page->dpaction.actions));
	gtk_clist_clear(GTK_CLIST(page->dpaction.actions));
	if((alist = stl_style_property_get_actions(stl)) != NULL)
	{
		gchar	*rd[2];
		gint	row;
		GList	*iter;

		for(iter = alist; iter != NULL; iter = g_list_next(iter))
		{
			rd[0] = iter->data;
			rd[1] = (gchar *) stl_style_property_get_action(stl, iter->data);
			row = gtk_clist_append(GTK_CLIST(page->dpaction.actions), rd);
		}
		g_list_free(alist);
	}
	gtk_clist_thaw(GTK_CLIST(page->dpaction.actions));
}

/* 1999-05-24 -	Set various editing widgets to display <stl>'s details. */
static void set_widgets(P_Styles *page, Style *stl)
{
	const gchar	*pname, *vpname[] = { SPN_COL_UNSEL_BG, SPN_COL_UNSEL_FG, SPN_ICON_UNSEL };
	gboolean	or;
	guint		i;

	if((page == NULL) || (stl == NULL))
		return;

	gtk_entry_set_text(GTK_ENTRY(page->dname), stl_style_get_name(stl));
	if((pname = stl_style_get_name(stl_styleinfo_style_get_parent(page->si, stl))) != NULL)
		gtk_entry_set_text(GTK_ENTRY(page->dparent), pname);
	else
		gtk_entry_set_text(GTK_ENTRY(page->dparent), _("(None)"));
	gtk_widget_set_sensitive(page->dreparent, pname != NULL);

	set_widgets_preview(page, stl);
	for(i = 0; i < 3; i++)
	{
		or = stl_style_property_get_override(stl, vpname[i]);
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(page->dpvisual.override[i]), or);
		gtk_widget_set_sensitive(page->dpvisual.override[i], (i < 1) || (pname != NULL));
		gtk_widget_set_sensitive(page->dpvisual.edit[i], or);
	}
	set_widgets_action(page, stl);
	gtk_widget_set_sensitive(page->dvbox, TRUE);
	gtk_widget_set_sensitive(page->del, stl_styleinfo_style_root(page->si) != stl);
}

/* 1999-05-25 -	Reset action editing widgets. */
static void reset_widgets_action(P_Styles *page)
{
	gtk_clist_unselect_all(GTK_CLIST(page->dpaction.actions));
	gtk_entry_set_text(GTK_ENTRY(page->dpaction.aname), "");
	gtk_widget_set_sensitive(page->dpaction.aname, FALSE);
	gtk_entry_set_text(GTK_ENTRY(page->dpaction.acmdseq), "");
	gtk_widget_set_sensitive(page->dpaction.acmdseq, FALSE);
	gtk_widget_set_sensitive(page->dpaction.acmdpick, FALSE);
	gtk_label_set_text(GTK_LABEL(GTK_BIN(page->dpaction.adel)->child), _("Delete Action"));
	gtk_widget_set_sensitive(page->dpaction.adel, FALSE);
}

/* 1999-05-24 -	Reset widgets. Handy when there is no longer a selection. */
static void reset_widgets(P_Styles *page)
{
	page->curr_prop  = NULL;
	page->curr_arow  = -1;
	gtk_entry_set_text(GTK_ENTRY(page->dname), "");
	gtk_entry_set_text(GTK_ENTRY(page->dparent), "");
	reset_widgets_action(page);
	gtk_widget_set_sensitive(page->dvbox, FALSE);
	gtk_widget_set_sensitive(page->del, FALSE);
}

/* ----------------------------------------------------------------------------------------- */

static Style * style_get_selected(P_Styles *page, GtkTreeIter *iter)
{
	GtkTreeIter	myiter;

	if(gtk_tree_selection_get_selected(gtk_tree_view_get_selection(GTK_TREE_VIEW(page->tree)), NULL, &myiter))
	{
		if(iter != NULL)
			*iter = myiter;
		return stl_styleinfo_get_style_iter(page->si, page->store, &myiter);
	}
	return NULL;
}

/* 2009-03-24 -	New-style selection tracking. */
static void evt_style_selection_changed(GtkTreeSelection *sel, gpointer user)
{
	P_Styles	*page = user;
	Style		*style;

	if((style = style_get_selected(page, NULL)) != NULL)
	{
		set_widgets(page, style);
		reset_widgets_action(page);
	}
	else
		reset_widgets(page);
}

/* 2009-02-06 -	A row was expanded, update the underlying Style object. */
static void evt_style_row_expanded(GtkTreeView *view, GtkTreeIter *iter, GtkTreePath *path, gpointer user_data)
{
	GtkTreeModel	*m;
	Style		*stl;

	m = gtk_tree_view_get_model(view);
	gtk_tree_model_get(m, iter, 1, &stl, -1);
	stl_style_set_expand(stl, TRUE);
}

/* 2009-02-06 -	A row was collapsed, update the underlying Style object. */
static void evt_style_row_collapsed(GtkTreeView *view, GtkTreeIter *iter, GtkTreePath *path, gpointer user_data)
{
	GtkTreeModel	*m;
	Style		*stl;

	m = gtk_tree_view_get_model(view);
	gtk_tree_model_get(m, iter, 1, &stl, -1);
	stl_style_set_expand(stl, FALSE);
}

/* 2009-02-06 -	This is a gtk_tree_model_foreach() callback, that simply applies the collapsed/
**		expanded status to a tree row showing a style. This is a property of the view,
**		so it cannot be done from inside the styles module.
*/
static gboolean set_expand(GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, gpointer user)
{
	Style	*stl;

	gtk_tree_model_get(model, iter, 1, &stl, -1);
	if(stl_style_get_expand(stl))
		gtk_tree_view_expand_row(GTK_TREE_VIEW(user), path, FALSE);
	else
		gtk_tree_view_collapse_row(GTK_TREE_VIEW(user), path);
	return FALSE;
}

/* 1999-05-24 -	Repopulate the tree. */
static void populate_tree(P_Styles *page)
{
	if(page->si != NULL)
	{
		page->store = stl_styleinfo_build_partial2(page->si, NULL);
		/* Block expand/collapse signals. */
		g_signal_handler_block(page->tree, page->sig_collapse);
		g_signal_handler_block(page->tree, page->sig_expand);
		gtk_tree_view_set_model(GTK_TREE_VIEW(page->tree), GTK_TREE_MODEL(page->store));
		gtk_tree_model_foreach(GTK_TREE_MODEL(page->store), set_expand, page->tree);
		g_signal_handler_unblock(page->tree, page->sig_expand);
		g_signal_handler_unblock(page->tree, page->sig_collapse);
	}
}

/* ----------------------------------------------------------------------------------------- */

/* 1999-05-24 -	Set a new name for the current style. */
static void evt_style_name_changed(GtkWidget *wid, gpointer user)
{
	P_Styles	*page = user;
	const gchar	*name;
	Style		*style;
	GtkTreeIter	iter;

	if((style = style_get_selected(page, &iter)) != NULL)
	{
		if((name = gtk_entry_get_text(GTK_ENTRY(wid))) != NULL)
		{
			ctp_replace_style(stl_style_get_name(style), style);
			stl_styleinfo_set_name_iter(page->si, page->store, &iter, name);
			page->modified = TRUE;
		}
	}
}

/* 1999-05-24 -	User clicked the details (magnifying glass) button to set new parent for current style. */
static void evt_style_parent_clicked(GtkWidget *wid, gpointer user)
{
	P_Styles	*page = user;
	Style		*style;
	Style		*np;

	if((style = style_get_selected(page, NULL)) != NULL)
	{
		if((np = sdl_dialog_sync_new_wait(page->si, style)) != NULL)
		{
			stl_styleinfo_style_set_parent(page->si, style, np);
			populate_tree(page);
			page->modified = TRUE;
		}
	}
}

/* ----------------------------------------------------------------------------------------- */

/* 1999-05-24 -	User toggled one of the visual property override checkbuttons. Update
**		the current style accordingly.
*/
static void evt_vprop_override_toggled(GtkWidget *wid, gpointer user)
{
	P_Styles	*page = user;
	Style		*style;
	const gchar	*pname;
	gboolean	or;

	if((style = style_get_selected(page, NULL)) == NULL)
		return;

	pname = gtk_object_get_user_data(GTK_OBJECT(wid));
	or = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(wid));
	if(or)
	{
		if(strcmp(pname, SPN_ICON_UNSEL) == 0)
			stl_style_property_set_icon(style, pname, stl_style_property_get_icon(style, pname));
		else
		{
			const GdkColor	*def;

			/* FIXME: This assumes only background can be non-overridden in root. */
			if((def = stl_style_property_get_color(style, pname)) == NULL)
				def = &(gtk_widget_get_style(GTK_WIDGET(page->min->gui->pane[0].list))->base[GTK_STATE_NORMAL]);
			stl_style_property_set_color(style, pname, def);
		}
	}
	else
		stl_style_property_remove(style, pname);
	set_widgets(page, style);
	page->modified = TRUE;
}

/* 1999-05-24 -	A color property has changed. Update previews and stuff. */
static void evt_vprop_color_changed(GdkColor *color, gpointer user)
{
	P_Styles	*page = user;
	Style		*style;

	if((style = style_get_selected(page, NULL)) == NULL)
		return;

	stl_style_property_set_color(style, page->curr_prop, color);
	set_widgets(page, style);
}

/* 1999-05-24 -	User hit the "Edit..." (or, for the icon, "Pick...") button below override toggle. */
static void evt_vprop_edit_clicked(GtkWidget *wid, gpointer user)
{
	P_Styles	*page = user;
	Style		*style;

	if((style = style_get_selected(page, NULL)) == NULL)
		return;

	page->curr_prop = gtk_object_get_user_data(GTK_OBJECT(wid));

	if(strcmp(page->curr_prop, SPN_ICON_UNSEL) == 0)
	{
		const gchar	*icon;

		if((icon = idl_dialog_sync_new_wait(page->min, cpt_get_path(PTID_ICON), NULL,
						  stl_style_property_get_icon(style, page->curr_prop),
							TRUE)) != NULL)
		{
			stl_style_property_set_icon(style, page->curr_prop, icon);
			page->modified = TRUE;
			set_widgets(page, style);
		}
	}
	else
	{
		const GdkColor	*col;
		GdkColor	initial;

		if((col = stl_style_property_get_color(style, page->curr_prop)) != NULL)
			initial = *col;
		if(cdl_dialog_sync_new_wait(_("Edit Color"), evt_vprop_color_changed, &initial, page) != DLG_POSITIVE)
			stl_style_property_set_color(style, page->curr_prop, &initial);
		else
			page->modified = TRUE;
		set_widgets(page, style);
	}
}

/* ----------------------------------------------------------------------------------------- */

/* 1999-05-25 -	Set the label of the action "Delete" button. */
static void action_set_delete_button(P_Styles *page)
{
	gboolean	unique, or;
	Style		*style;

	if((style = style_get_selected(page, NULL)) == NULL)
		return;

	unique = stl_style_property_is_unique(style, page->curr_prop);
	gtk_label_set_text(GTK_LABEL(GTK_BIN(page->dpaction.adel)->child), unique ? _("Delete Action") :
				_("Revert to Inherited Command"));
	or = stl_style_property_get_override(style, page->curr_prop);
	gtk_entry_set_editable(GTK_ENTRY(page->dpaction.aname), or);
	gtk_widget_set_sensitive(page->dpaction.adel, or);
}

/* 1999-05-25 -	An action property has been selected. Update label on "delete/override" button. */
static void evt_aprop_selected(GtkWidget *wid, gint row, gint column, GdkEventButton *evt, gpointer user)
{
	P_Styles	*page = user;
	Style		*style;
	gchar		*prop;

	if((style = style_get_selected(page, NULL)) == NULL)
		return;

	page->curr_prop = NULL;
	page->curr_arow = -1;
	gtk_clist_get_text(GTK_CLIST(page->dpaction.actions), row, 0, (gchar **) &prop);
	gtk_entry_set_text(GTK_ENTRY(page->dpaction.aname), prop);
	gtk_widget_set_sensitive(page->dpaction.aname, TRUE);
	gtk_entry_set_text(GTK_ENTRY(page->dpaction.acmdseq), stl_style_property_get_action(style, prop));
	/* Force execution of cmdseq_changed() before we set curr_prop & curr_arow. */
	gui_events_flush();
	gtk_widget_set_sensitive(page->dpaction.acmdseq, TRUE);
	gtk_widget_set_sensitive(page->dpaction.acmdpick, TRUE);
	page->curr_prop = prop;
	page->curr_arow = row;
	action_set_delete_button(page);
}

/* 1999-05-25 -	An action was just unselected. Update widgets. */
static void evt_aprop_unselected(GtkWidget *wid, gint row, gint column, GdkEventButton *evt, gpointer user)
{
	P_Styles	*page = user;
	Style		*style;

	if((style = style_get_selected(page, NULL)) == NULL)
		return;

	page->curr_prop = NULL;
	page->curr_arow = -1;
	reset_widgets_action(page);
}

/* 2004-02-24 -	Action was renamed. */
static void evt_aprop_name_changed(GtkWidget *wid, gpointer user)
{
	P_Styles	*page = user;
	Style		*style;

	if((style = style_get_selected(page, NULL)) == NULL)
		return;
	if((page->curr_prop != NULL) && (page->curr_arow >= 0))
	{
		gchar		*oname;
		const gchar	*nname;

		gtk_clist_get_text(GTK_CLIST(page->dpaction.actions), page->curr_arow, 0, &oname);
		if((oname != NULL) && (nname = gtk_entry_get_text(GTK_ENTRY(wid))) != NULL)
		{
			if(stl_style_property_rename(style, oname, nname))	/* Will fail on collision. */
				gtk_clist_set_text(GTK_CLIST(page->dpaction.actions), page->curr_arow, 0, nname);
		}
	}
}

/* 1999-05-25 -	Command sequence of current action property changed. Register that. */
static void evt_aprop_cmdseq_changed(GtkWidget *wid, gpointer user)
{
	P_Styles	*page = user;
	Style		*style;

	if((style = style_get_selected(page, NULL)) == NULL)
		return;

	if((page->curr_prop != NULL) && (page->curr_arow >= 0))
	{
		const gchar	*text;

		if((text = gtk_entry_get_text(GTK_ENTRY(wid))) != NULL)
		{
			gtk_clist_set_text(GTK_CLIST(page->dpaction.actions), page->curr_arow, 1, text);
			stl_style_property_set_action(style, page->curr_prop, text);
			action_set_delete_button(page);
			page->modified = TRUE;
		}
	}
}

/* 1999-05-25 -	User hit the details (magnifying glass) button to the right of the Command field. Dialog-time. */
static void evt_aprop_cmdpick_clicked(GtkWidget *wid, gpointer user)
{
	P_Styles	*page = user;
	Style		*style;

	if((style = style_get_selected(page, NULL)) == NULL)
		return;

	if((page->curr_prop != NULL) && (page->curr_arow >= 0))
	{
		const gchar	*cmd;

		if((cmd = csq_dialog_sync_new_wait(page->min, ccs_get_current())) != NULL)
			gtk_entry_set_text(GTK_ENTRY(page->dpaction.acmdseq), cmd);
	}
}

/* 1999-05-25 -	Add a new action property. Pops up a dialog asking for the name. */
static void evt_aprop_add_clicked(GtkWidget *wid, gpointer user)
{
	P_Styles	*page = user;
	Style		*style;
	Dialog		*dlg;
	GtkWidget	*hbox, *label, *entry;

	if((style = style_get_selected(page, NULL)) == NULL)
		return;

	hbox = gtk_hbox_new(FALSE, 0);
	label = gtk_label_new(_("Name"));
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
	entry = gtk_entry_new_with_max_length(STL_PROPERTY_NAME_SIZE - 1);
	gtk_box_pack_start(GTK_BOX(hbox), entry, TRUE, TRUE, 0);
	dlg = dlg_dialog_sync_new(hbox, _("New Action Property"), NULL);
	gtk_widget_grab_focus(entry);
	if(dlg_dialog_sync_wait(dlg) == DLG_POSITIVE)
	{
		const gchar	*name;

		if(((name = gtk_entry_get_text(GTK_ENTRY(entry))) != NULL) && *name != '\0')
		{
			stl_style_property_set_action(style, name, _("something"));
			set_widgets_action(page, style);
			reset_widgets_action(page);
			page->modified = TRUE;
		}
	}
	dlg_dialog_sync_destroy(dlg);
}

/* 1999-05-25 -	User clicked the "Delete" (or "Revert...") button. Do it. */
static void evt_aprop_delete_clicked(GtkWidget *wid, gpointer user)
{
	P_Styles	*page = user;
	Style		*style;

	if((style = style_get_selected(page, NULL)) == NULL)
		return;

	if((page->curr_prop != NULL) && (page->curr_arow >= 0))
	{
		stl_style_property_remove(style, page->curr_prop);
		gtk_clist_unselect_all(GTK_CLIST(page->dpaction.actions));
		set_widgets_action(page, style);
		reset_widgets_action(page);
		page->modified = TRUE;
	}
}

/* ----------------------------------------------------------------------------------------- */

/* 1999-05-26 -	Add a new style. Use the currently selected style, if one exists, as parent.
**		If there is no selection, add style with Root as parent.
*/
static void evt_style_add_clicked(GtkWidget *wid, gpointer user)
{
	P_Styles	*page = user;

	if(page != NULL)
	{
		Style		*stl, *parent;
		GtkTreeIter	iter;

		stl = stl_style_new_unique_name(page->si);
		/* Parent is either current selection, or root. */
		if((parent = style_get_selected(page, &iter)) == NULL)
			parent = stl_styleinfo_style_root(page->si);
		stl_styleinfo_style_add(page->si, parent, stl);
		stl_style_set_expand(parent, TRUE);
		populate_tree(page);
		if(stl_styleinfo_tree_find_style(page->si, page->store, stl, &iter))
		{
			gtk_tree_selection_select_iter(gtk_tree_view_get_selection(GTK_TREE_VIEW(page->tree)), &iter);
			gtk_entry_select_region(GTK_ENTRY(page->dname), 0, -1);
			gtk_widget_grab_focus(page->dname);
		}
	}
}

/* 1999-05-27 -	Delete the currently selected style. If the style has children, the user must
**		confirm the operation, since all children will be deleted, too.
*/
static void evt_style_del_clicked(GtkWidget *wid, gpointer user)
{
	P_Styles	*page = user;
	Style		*style;
	gint		ok = DLG_POSITIVE;

	if((style = style_get_selected(page, NULL)) == NULL)
		return;

	if(stl_styleinfo_style_has_children(page->si, style))
		ok = dlg_dialog_sync_new_simple_wait(_("Deleting this style will also delete\n"
			"all its children. Are you sure?"), _("Confirm Delete"), _("_Delete|_Cancel"));
	if(ok == DLG_POSITIVE)
	{
		GList	*chlist, *iter;
		Style	*stl;

		stl = stl_styleinfo_style_root(page->si);
		chlist = stl_styleinfo_style_get_children(page->si, style, TRUE);
		for(iter = chlist; iter != NULL; iter = g_list_next(iter))
			ctp_replace_style(stl_style_get_name(iter->data), stl);
		g_list_free(chlist);
		stl_styleinfo_style_remove(page->si, style);
		populate_tree(page);
		reset_widgets(page);
	}
}

/* ----------------------------------------------------------------------------------------- */

/* 1999-05-22 -	Build the visual property editing widgets. */
static void build_pvisual(P_Styles *page)
{
	PVisual		*pv = &page->dpvisual;
	const gchar	*vplab[] = { N_("Background Color"), N_("Foreground Color"), N_("Icon") },
			*vpname[] = { SPN_COL_UNSEL_BG, SPN_COL_UNSEL_FG, SPN_ICON_UNSEL };
	gchar		*ctxt[] = { "", N_("(Row Style Preview Text)") };
	GtkWidget	*hbox, *frame, *vbox, *label;
	guint		i;

	pv->vbox = gtk_vbox_new(FALSE, 0);
	hbox = gtk_hbox_new(FALSE, 0);
	label = gtk_label_new(_("Preview"));
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
	pv->preview = gtk_clist_new(2);
	gtk_widget_set_name(pv->preview, "cstPreview");
	gtk_clist_set_column_width(GTK_CLIST(pv->preview), 0, 16);
	ctxt[1] = _(ctxt[1]);
	gtk_clist_append(GTK_CLIST(pv->preview), ctxt);
	gtk_clist_set_selectable(GTK_CLIST(pv->preview), 0, FALSE);
	gtk_box_pack_start(GTK_BOX(hbox), pv->preview, TRUE, TRUE, 0);
	gtk_box_pack_start(GTK_BOX(pv->vbox), hbox, FALSE, FALSE, 2);

	hbox = gtk_hbox_new(FALSE, 0);
	for(i = 0; i < 3; i++)
	{
		frame = gtk_frame_new(_(vplab[i]));
		vbox = gtk_vbox_new(FALSE, 0);
		gtk_container_set_border_width(GTK_CONTAINER(vbox), 5);
		pv->override[i] = gtk_check_button_new_with_label(_("Override Parent's?"));
		gtk_object_set_user_data(GTK_OBJECT(pv->override[i]), (gpointer) vpname[i]);
		g_signal_connect(G_OBJECT(pv->override[i]), "toggled", G_CALLBACK(evt_vprop_override_toggled), page);
		gtk_box_pack_start(GTK_BOX(vbox), pv->override[i], FALSE, FALSE, 0);
		pv->edit[i] = gtk_button_new_with_label(i < 2 ? _("Edit...") : _("Pick..."));
		gtk_object_set_user_data(GTK_OBJECT(pv->edit[i]), (gpointer) vpname[i]);
		g_signal_connect(G_OBJECT(pv->edit[i]), "clicked", G_CALLBACK(evt_vprop_edit_clicked), page);
		gtk_box_pack_start(GTK_BOX(vbox), pv->edit[i], FALSE, FALSE, 0);
		gtk_container_add(GTK_CONTAINER(frame), vbox);
		gtk_box_pack_start(GTK_BOX(hbox), frame, TRUE, TRUE, 5);
	}
	gtk_box_pack_start(GTK_BOX(pv->vbox), hbox, FALSE, FALSE, 0);
}

/* 1999-05-22 -	Build action property editing widgetry. */
static void build_paction(P_Styles *page)
{
	gchar		*at[] = { N_("Name"), N_("Command") };
	PAction		*pa = &page->dpaction;
	GtkWidget	*hbox, *btn, *table, *label, *scwin;

	pa->vbox = gtk_vbox_new(FALSE, 0);
	scwin = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scwin), GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);
	pa->actions = gtk_clist_new(2);
	gtk_clist_set_column_justification(GTK_CLIST(pa->actions), 0, GTK_JUSTIFY_LEFT);
	gtk_clist_set_column_justification(GTK_CLIST(pa->actions), 1, GTK_JUSTIFY_LEFT);
	gtk_clist_set_column_auto_resize(GTK_CLIST(pa->actions), 0, TRUE);
	gtk_clist_set_column_min_width(GTK_CLIST(pa->actions), 0, 128);
	gtk_clist_set_column_resizeable(GTK_CLIST(pa->actions), 1, FALSE);
	gtk_clist_append(GTK_CLIST(pa->actions), at);
	gtk_clist_append(GTK_CLIST(pa->actions), at);
	gtk_clist_append(GTK_CLIST(pa->actions), at);
	gtk_clist_append(GTK_CLIST(pa->actions), at);
	gtk_clist_append(GTK_CLIST(pa->actions), at);
	gtk_clist_append(GTK_CLIST(pa->actions), at);
	gtk_clist_append(GTK_CLIST(pa->actions), at);
	gtk_clist_append(GTK_CLIST(pa->actions), at);
	g_signal_connect(G_OBJECT(pa->actions), "select_row", G_CALLBACK(evt_aprop_selected), page);
	g_signal_connect(G_OBJECT(pa->actions), "unselect_row", G_CALLBACK(evt_aprop_unselected), page);
	gtk_container_add(GTK_CONTAINER(scwin), pa->actions);
	gtk_box_pack_start(GTK_BOX(pa->vbox), scwin, TRUE, TRUE, 0);

	table = gtk_table_new(2, 3, FALSE);
	label = gtk_label_new(_("Name"));
	gtk_table_attach(GTK_TABLE(table), label, 0, 1, 0, 1,  0,0,0,0);
	pa->aname = gtk_entry_new_with_max_length(STL_PROPERTY_NAME_SIZE - 1);
	gtk_entry_set_editable(GTK_ENTRY(pa->aname), FALSE);
	g_signal_connect(G_OBJECT(pa->aname), "changed", G_CALLBACK(evt_aprop_name_changed), page);
	gtk_table_attach(GTK_TABLE(table), pa->aname, 1, 3, 0, 1,  GTK_EXPAND|GTK_FILL,0,0,0);
	label = gtk_label_new(_("Command"));
	gtk_table_attach(GTK_TABLE(table), label, 0, 1, 1, 2,  0,0,0,0);
	pa->acmdseq = gtk_entry_new();
	g_signal_connect(G_OBJECT(pa->acmdseq), "changed", G_CALLBACK(evt_aprop_cmdseq_changed), page);
	gtk_table_attach(GTK_TABLE(table), pa->acmdseq, 1, 2, 1, 2,  GTK_EXPAND|GTK_FILL,0,0,0);
	pa->acmdpick = gui_details_button_new(page->min->gui->window->window);
	g_signal_connect(G_OBJECT(pa->acmdpick), "clicked", G_CALLBACK(evt_aprop_cmdpick_clicked), page);
	gtk_table_attach(GTK_TABLE(table), pa->acmdpick, 2, 3, 1, 2,  0,0,0,0);
	gtk_box_pack_start(GTK_BOX(pa->vbox), table, FALSE, FALSE, 0);

	hbox = gtk_hbox_new(FALSE, 0);
	btn = gtk_button_new_with_label(_("Add Action..."));
	g_signal_connect(G_OBJECT(btn), "clicked", G_CALLBACK(evt_aprop_add_clicked), page);
	gtk_box_pack_start(GTK_BOX(hbox), btn, TRUE, TRUE, 5);
	pa->adel = gtk_button_new_with_label(_("Delete Action"));
	g_signal_connect(G_OBJECT(pa->adel), "clicked", G_CALLBACK(evt_aprop_delete_clicked), page);
	gtk_box_pack_start(GTK_BOX(hbox), pa->adel, TRUE, TRUE, 5);
	gtk_box_pack_start(GTK_BOX(pa->vbox), hbox, FALSE, FALSE, 5);
}

/* 1999-05-22 -	Build style configuration GUI page. */
static GtkWidget * cst_init(MainInfo *min, gchar **name)
{
	P_Styles	*page = &the_page;
	GtkWidget	*table, *label, *frame, *hbox, *btn;
	GtkCellRenderer	*cr;
	GtkTreeViewColumn *vc;

	page->min = min;
	page->si  = NULL;
	page->modified = FALSE;
	page->store = NULL;

	page->vbox = gtk_vbox_new(FALSE, 0);
	page->scwin = gtk_scrolled_window_new(FALSE, FALSE);
	page->tree = gtk_tree_view_new();
	cr = gtk_cell_renderer_text_new();
	vc = gtk_tree_view_column_new_with_attributes("(Styles)", cr, "text", 0, NULL);
	gtk_tree_view_append_column(GTK_TREE_VIEW(page->tree), vc);
	gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(page->tree), FALSE);
	g_signal_connect(G_OBJECT(gtk_tree_view_get_selection(GTK_TREE_VIEW(page->tree))), "changed", G_CALLBACK(evt_style_selection_changed), page);
	page->sig_expand   = g_signal_connect(G_OBJECT(page->tree), "row_expanded", G_CALLBACK(evt_style_row_expanded), page);
	page->sig_collapse = g_signal_connect(G_OBJECT(page->tree), "row_collapsed", G_CALLBACK(evt_style_row_collapsed), page);
	gtk_container_add(GTK_CONTAINER(page->scwin), page->tree);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(page->scwin), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_box_pack_start(GTK_BOX(page->vbox), page->scwin, TRUE, TRUE, 0);

	page->dvbox = gtk_vbox_new(FALSE, 0);

	table = gtk_table_new(2, 3, FALSE);
	label = gtk_label_new(_("Name"));
	gtk_table_attach(GTK_TABLE(table), label, 0, 1, 0, 1,  0,0,0,0);
	page->dname = gtk_entry_new_with_max_length(STL_STYLE_NAME_SIZE - 1);
	g_signal_connect(G_OBJECT(page->dname), "changed", G_CALLBACK(evt_style_name_changed), page);
	gtk_table_attach(GTK_TABLE(table), page->dname, 1, 3, 0, 1,  GTK_EXPAND|GTK_FILL,0,0,0);
	label = gtk_label_new(_("Parent"));
	gtk_table_attach(GTK_TABLE(table), label, 0, 1, 1, 2,  0,0,0,0);
	page->dparent = gtk_entry_new_with_max_length(STL_STYLE_NAME_SIZE - 1);
	gtk_entry_set_editable(GTK_ENTRY(page->dparent), FALSE);
	gtk_table_attach(GTK_TABLE(table), page->dparent, 1, 2, 1, 2,  GTK_EXPAND|GTK_FILL,0,0,0);
	page->dreparent = gui_details_button_new(min->gui->window->window);
	g_signal_connect(G_OBJECT(page->dreparent), "clicked", G_CALLBACK(evt_style_parent_clicked), page);
	gtk_table_attach(GTK_TABLE(table), page->dreparent, 2, 3, 1, 2,  0,0,0,0);
	gtk_box_pack_start(GTK_BOX(page->dvbox), table, FALSE, FALSE, 0);

	frame = gtk_frame_new(_("Inherited Properties"));
	page->dpnbook = gtk_notebook_new();
	gtk_notebook_set_tab_pos(GTK_NOTEBOOK(page->dpnbook), GTK_POS_LEFT);
	build_pvisual(page);
	gtk_notebook_append_page(GTK_NOTEBOOK(page->dpnbook), page->dpvisual.vbox, gtk_label_new(_("Visual")));
	build_paction(page);
	gtk_notebook_append_page(GTK_NOTEBOOK(page->dpnbook), page->dpaction.vbox, gtk_label_new(_("Actions")));
	gtk_container_add(GTK_CONTAINER(frame), page->dpnbook);

	gtk_box_pack_start(GTK_BOX(page->dvbox), frame, FALSE, FALSE, 5);
	gtk_box_pack_start(GTK_BOX(page->vbox), page->dvbox, FALSE, FALSE, 0);

	hbox = gtk_hbox_new(FALSE, 0);
	btn = gtk_button_new_with_label(_("Add"));
	g_signal_connect(G_OBJECT(btn), "clicked", G_CALLBACK(evt_style_add_clicked), page);
	gtk_box_pack_start(GTK_BOX(hbox), btn, TRUE, TRUE, 5);
	page->del = gtk_button_new_with_label(_("Delete"));
	g_signal_connect(G_OBJECT(page->del), "clicked", G_CALLBACK(evt_style_del_clicked), page);
	gtk_box_pack_start(GTK_BOX(hbox), page->del, TRUE, TRUE, 5);
	gtk_box_pack_start(GTK_BOX(page->vbox), hbox, FALSE, FALSE, 5);

	gtk_widget_show_all(page->vbox);

	cfg_tree_level_begin(_("File Recognition"));
	cfg_tree_level_append(_("Styles"), page->vbox);

	return NULL;
}

/* ----------------------------------------------------------------------------------------- */

/* 1999-05-12 -	Update style display. */
static void cst_update(MainInfo *min)
{
	P_Styles	*page = &the_page;

	page->si = stl_styleinfo_copy(min->cfg.style);
	populate_tree(page);
	reset_widgets(page);
}

/* ----------------------------------------------------------------------------------------- */

/* 1999-05-27 -	Accept changes, if any. Causes a call to the relink function in types config.
** 1999-06-13 -	Fixed a rather sneaky problem: if the style for a _type_ changed, it would point
**		into the editing copies here. If the editing copies went away because of the
**		'modified' flag being FALSE, those types were left with dangling style pointers.
**		The fix was simple.
*/
static void cst_accept(MainInfo *min)
{
	P_Styles	*page = &the_page;

	if(page->modified)
	{
		ctp_relink_styles(min->cfg.style, page->si);
		stl_styleinfo_destroy(min->cfg.style);
		min->cfg.style = page->si;
		page->si = NULL;
		page->modified = FALSE;
		cfg_set_flags(CFLG_RESCAN_LEFT | CFLG_RESCAN_RIGHT);
	}
	else	/* Make sure the types link to the existing styles. */
		ctp_relink_styles(page->si, min->cfg.style);
}

/* ----------------------------------------------------------------------------------------- */

/* 1999-05-27 -	Save the current style configuration. Thanks to the styles module, this
**		is really not complicated. :)
*/
static gint cst_save(MainInfo *min, FILE *out)
{
	stl_styleinfo_save(min, min->cfg.style, out, NODE);

	return TRUE;
}

/* 1998-08-24 -	Load style configuration info from given XML tree. */
static void cst_load(MainInfo *min, const XmlNode *node)
{
	/* Free the built-in first. */
	stl_styleinfo_destroy(min->cfg.style);
	min->cfg.style = stl_styleinfo_load(node);
}

/* ----------------------------------------------------------------------------------------- */

/* 1999-05-27 -	When the config window closes, free the editing copies since they're bulky. */
static void cst_hide(MainInfo *min)
{
	P_Styles	*page = &the_page;

	stl_styleinfo_destroy(page->si);
	page->si = NULL;
	populate_tree(page);	/* This will destroy the tree widget and *not* create a new one, since there's no si. */
	reset_widgets(page);
}

/* ----------------------------------------------------------------------------------------- */

const CfgModule * cst_describe(MainInfo *min)
{
	static const CfgModule	desc = { NODE, cst_init, cst_update, cst_accept, cst_save, cst_load, cst_hide };

	return &desc;
}

/* ----------------------------------------------------------------------------------------- */

/* 1999-05-27 -	Get the most current style info, namely the editing version. */
StyleInfo * cst_get_styleinfo(void)
{
	return the_page.si;
}
